# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository Tree Items (RTIs) for netCDF data.

    It uses the netCDF4 package to open netCDF files.
    See http://unidata.github.io/netcdf4-python/
"""

import logging, types
from netCDF4 import Dataset, Variable, Dimension

from argos.utils.cls import check_class
from argos.repo.baserti import BaseRti
from argos.repo.iconfactory import RtiIconFactory

logger = logging.getLogger(__name__)

ICON_COLOR_NCDF4 = '#0088FF'


def ncVarAttributes(ncVar):
    """ Returns the attributes of ncdf variable
    """
    try:
        return ncVar.__dict__
    except Exception as ex:
        # Due to some internal error netCDF4 may raise an AttributeError or KeyError,
        # depending on its version.
        logger.warn("Unable to read the attributes from {}. Reason: {}"
                    .format(ncVar.name, ex))
        return {}


def ncVarUnit(ncVar):
    """ Returns the unit of the ncVar by looking in the attributes.

        It searches in the attributes for one of the following keys:
        'unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'. If these are not found, the empty
        string is returned.
    """
    attributes = ncVarAttributes(ncVar)
    if not attributes:
        return '' # a premature optimization :-)

    for key in ('unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'):
        if key in attributes:
            # In Python3 the attribures are byte strings so we must decode them
            # This a bug in h5py, see https://github.com/h5py/h5py/issues/379
            return attributes[key]
    else:
        return ''



def variableMissingValue(ncVar):
    """ Returns the missingData given a NetCDF variable

        Looks for one of the following attributes: _FillValue, missing_value, MissingValue,
        missingValue. Returns None if these attributes are not found.
    """
    attributes = ncVarAttributes(ncVar)
    if not attributes:
        return None # a premature optimization :-)

    for key in ('missing_value', 'MissingValue', 'missingValue', 'FillValue', '_FillValue'):
        if key in attributes:
            missingDataValue = attributes[key]
            return missingDataValue
    return None



class NcdfDimensionRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a NCDF group.
    """
    _defaultIconGlyph = RtiIconFactory.DIMENSION
    _defaultIconColor = ICON_COLOR_NCDF4

    def __init__(self, ncDim, nodeName, fileName=''):
        """ Constructor
        """
        super(NcdfDimensionRti, self).__init__(nodeName, fileName=fileName)
        check_class(ncDim, Dimension)

        self._ncDim = ncDim

    def hasChildren(self):
        """ Returns False. Dimension items never have children.
        """
        return False

    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return {'unlimited': str(self._ncDim.isunlimited())}
        #size = self._ncDim.size
        #return {'size': 'unlimited' if size is None else str(size)}



class NcdfFieldRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a field in a structured NCDF variable.
    """
    _defaultIconGlyph = RtiIconFactory.FIELD
    _defaultIconColor = ICON_COLOR_NCDF4

    def __init__(self, ncVar, nodeName, fileName=''):
        """ Constructor.
            The name of the field must be given to the nodeName parameter.
        """
        super(NcdfFieldRti, self).__init__(nodeName, fileName=fileName)
        check_class(ncVar, Variable)

        self._ncVar = ncVar

    def hasChildren(self):
        """ Returns False. Field items never have children.
        """
        return False


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Applies the index on the NCDF variable that contain this field and then selects the
            current field. In pseudo-code, it returns: self.h5Dataset[index][self.nodeName].

            If the field itself contains a sub-array it returns:
                self.h5Dataset[mainArrayIndex][self.nodeName][subArrayIndex]
        """
        mainArrayNumDims = self._ncVar.ndim
        mainIndex = index[:mainArrayNumDims]
        mainArray = self._ncVar.__getitem__(mainIndex)
        fieldArray = mainArray[self.nodeName]
        subIndex = tuple([Ellipsis]) + index[mainArrayNumDims:]
        slicedArray = fieldArray[subIndex]
        return slicedArray


    @property
    def nDims(self):
        """ The number of dimensions of the underlying array
        """
        return self._ncVar.ndim + len(self._subArrayShape)


    @property
    def _subArrayShape(self):
        """ Returns the shape of the sub-array
            An empty tuple is returned for regular fields, which have no sub array.
        """
        if self._ncVar.dtype.fields is None:
            return tuple() # regular field # TODO: does this occur?
        else:
            fieldName = self.nodeName
            fieldDtype = self._ncVar.dtype.fields[fieldName][0]
            return fieldDtype.shape


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
            If the field contains a subarray the shape may be longer than 1.
        """
        return self._ncVar.shape + self._subArrayShape


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        fieldName = self.nodeName
        return str(self._ncVar.dtype.fields[fieldName][0])


    @property
    def attributes(self):
        """ The attributes dictionary.
            Returns the attributes of the variable that contains this field.
        """
        return ncVarAttributes(self._ncVar)


    @property
    def unit(self):
        """ Returns the unit attribute of the underlying ncdf variable.

            If the units has a length (e.g is a list) and has precisely one element per field,
            the unit for this field is returned.
        """
        unit = ncVarUnit(self._ncVar)
        fieldNames = self._ncVar.dtype.names

        # If the missing value attribute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(unit, '__len__') and len(unit) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return unit[idx]
        else:
            return unit



    @property
    def dimensionNames(self):
        """ Returns a list with the dimension names of the underlying NCDF variable
        """
        nSubDims = len(self._subArrayShape)
        subArrayDims = ['SubDim{}'.format(dimNr) for dimNr in range(nSubDims)]
        return list(self._ncVar.dimensions + tuple(subArrayDims))


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        value = variableMissingValue(self._ncVar)
        fieldNames = self._ncVar.dtype.names

        # If the missing value attibute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(value, '__len__') and len(value) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return value[idx]
        else:
            return value


class NcdfVariableRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a NCDF variable.
    """
    _defaultIconGlyph = RtiIconFactory.ARRAY
    _defaultIconColor = ICON_COLOR_NCDF4

    def __init__(self, ncVar, nodeName, fileName=''):
        """ Constructor
        """
        super(NcdfVariableRti, self).__init__(nodeName, fileName=fileName)
        check_class(ncVar, Variable)
        self._ncVar = ncVar

        try:
            self._isStructured = bool(self._ncVar.dtype.names)
        except (AttributeError, KeyError):
            # If dtype is a string instead of an numpy dtype, netCDF4 raises a KeyError
            # or AttributeError, depending on its version.
            self._isStructured = False

    def hasChildren(self):
        """ Returns True if the variable has a structured type, otherwise returns False.
        """
        return self._isStructured


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Passes the index through to the underlying array.
        """
        return self._ncVar.__getitem__(index)


    @property
    def nDims(self):
        """ The number of dimensions of the underlying array
        """
        return self._ncVar.ndim


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
        """
        return self._ncVar.shape


    @property
    def attributes(self):
        """ The attributes dictionary.
            Returns the attributes of the variable that contains this field.
        """
        return ncVarAttributes(self._ncVar)


    @property
    def unit(self):
        """ Returns the unit attribute of the underlying ncdf variable
        """
        return ncVarUnit(self._ncVar)


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        dtype =  self._ncVar.dtype
        if type(dtype) == type:
            # Handle the unexpected case that dtype is a regular Python type
            # (happens e.g. in the /PROCESSOR/processing_configuration of the Trop LX files)
            return dtype.__name__

        return '<structured>' if dtype.names else str(dtype)


    @property
    def dimensionNames(self):
        """ Returns a list with the dimension names of the underlying NCDF variable
        """
        return self._ncVar.dimensions

#    TODO: how to get this?
#    @property
#    def dimensionGroupPaths(self):
#        """ Returns a list with, for every dimension, the path of the group that contains it.
#        """
#        return [dim.group().path for dim in self._ncVar.dimensions.values()] # TODO: cache?
#

    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        return variableMissingValue(self._ncVar)


    def _fetchAllChildren(self):
        """ Fetches all fields that this variable contains.
            Only variables with a structured data type can have fields.
        """
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        # Add fields
        if self._isStructured:
            for fieldName in self._ncVar.dtype.names:
                childItems.append(NcdfFieldRti(self._ncVar, nodeName=fieldName, fileName=self.fileName))

        return childItems



class NcdfGroupRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a NCDF group.
    """
    _defaultIconGlyph = RtiIconFactory.FOLDER
    _defaultIconColor = ICON_COLOR_NCDF4

    def __init__(self, ncGroup, nodeName, fileName=''):
        """ Constructor
        """
        super(NcdfGroupRti, self).__init__(nodeName, fileName=fileName)
        check_class(ncGroup, Dataset, allow_none=True)

        self._ncGroup = ncGroup


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return self._ncGroup.__dict__ if self._ncGroup else {}


    def _fetchAllChildren(self):
        """ Fetches all sub groups and variables that this group contains.
        """
        assert self._ncGroup is not None, "dataset undefined (file not opened?)"
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        # Add dimensions
        for dimName, ncDim in self._ncGroup.dimensions.items():
            childItems.append(NcdfDimensionRti(ncDim, nodeName=dimName, fileName=self.fileName))

        # Add groups
        for groupName, ncGroup in self._ncGroup.groups.items():
            childItems.append(NcdfGroupRti(ncGroup, nodeName=groupName, fileName=self.fileName))

        # Add variables
        for varName, ncVar in self._ncGroup.variables.items():
            childItems.append(NcdfVariableRti(ncVar, nodeName=varName, fileName=self.fileName))

        return childItems



class NcdfFileRti(NcdfGroupRti):
    """ Reads a NetCDF file using the netCDF4 module.

        See http://unidata.github.io/netcdf4-python/
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_NCDF4

    def __init__(self, nodeName, fileName=''):
        """ Constructor
        """
        super(NcdfFileRti, self).__init__(None, nodeName, fileName=fileName)
        self._checkFileExists()

    def _openResources(self):
        """ Opens the root Dataset.
        """
        logger.info("Opening: {}".format(self._fileName))
        self._ncGroup = Dataset(self._fileName)

    def _closeResources(self):
        """ Closes the root Dataset.
        """
        logger.info("Closing: {}".format(self._fileName))
        self._ncGroup.close()
        self._ncGroup = None



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Stores for representing data that is read from text files.
"""
import logging, os
import numpy as np

from numpy.lib.npyio import NpzFile

from argos.qt import QtWidgets
from argos.repo.iconfactory import RtiIconFactory
from argos.repo.memoryrtis import ArrayRti, SliceRti, MappingRti
from argos.utils.cls import check_is_an_array, check_class

logger = logging.getLogger(__name__)

ICON_COLOR_NUMPY = '#987456'

# Do not allow pickle in numpy.load(), at least for now. This can be a security risk
ALLOW_PICKLE = False



class NumpyTextFileRti(ArrayRti):
    """ Reads a 2D array from a simple text file using numpy.loadtxt().
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_NUMPY

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(NumpyTextFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                               iconColor=self._defaultIconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file
        """
        self._array = np.loadtxt(self._fileName, ndmin=0)


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None


    def _fetchAllChildren(self):
        """ Adds an ArrayRti per column as children so that they can be inspected easily
        """
        childItems = []
        _nRows, nCols = self._array.shape if self._array is not None else (0, 0)
        for col in range(nCols):
            colItem = SliceRti(self._array[:, col], nodeName="column-{}".format(col),
                               fileName=self.fileName, iconColor=self.iconColor,
                               attributes=self.attributes)
            childItems.append(colItem)
        return childItems




class NumpyBinaryFileRti(ArrayRti):
    """ Reads a Numpy array from a binary file (.npy) using numpy.load().

        The file must have been saved with numpy.save() and therefore contain a single arrays.
        A TypeError is raised if this is not the case.

        The allow_pickle is set to False, no object arrays can be read.
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_NUMPY

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(NumpyBinaryFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                                 iconColor=self._defaultIconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children

            Returns True so that the file can be opened, even though the array has no children.
        """
        return True


    def _openResources(self):
        """ Uses numpy.load to open the underlying file
        """
        arr = np.load(self._fileName, allow_pickle=ALLOW_PICKLE)
        check_is_an_array(arr)
        self._array = arr


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None



class NumpyCompressedFileRti(MappingRti):
    """ Reads arrays from a Numpy zip file (.npz) using numpy.load().

        The file must have been saved with numpy.savez() and therefore contain multiple arrays.
        A TypeError is raised if this is not the case.

        The allow_pickle is set to False, no object arrays can be read.
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_NUMPY

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an MappingRti with None as underlying dictionary.
        """
        super(NumpyCompressedFileRti, self).__init__(None,
                                                     nodeName=nodeName, fileName=fileName,
                                                     iconColor=self._defaultIconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.load to open the underlying file
        """
        dct = np.load(self._fileName, allow_pickle=ALLOW_PICKLE)
        check_class(dct, NpzFile)
        self._dictionary = dct


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._dictionary = None




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Qt widgets specific to this project.

    Note that some widgets are locate elsewhere if it makes sense to place them near related
    classes. For example the ConfigTreeView widgets can be found in argos.config.configtreeview.
"""



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository tree items that are read using import routines of Scipy
    See http://docs.scipy.org/doc/scipy-0.16.0/reference/io.html
"""
import logging, os
import scipy
import scipy.io
import scipy.io.wavfile

from argos.repo.memoryrtis import ArrayRti, SliceRti, MappingRti
from argos.repo.iconfactory import RtiIconFactory

logger = logging.getLogger(__name__)

ICON_COLOR_SCIPY = '#987456'


class MatlabFileRti(MappingRti):
    """ Read data from an MATLAB file

        Uses scipy.io.loadmat to read the file.

        Note: v4 (Level 1.0), v6 and v7 to 7.2 matfiles are supported.

        You will need an HDF5 python library to read matlab 7.3 format mat files, because SciPy
        does not supply one; they do not implement the HDF5 / 7.3 interface.
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_SCIPY

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an MappingRti with None as underlying dictionary.
        """
        super(MatlabFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                            iconColor=self._defaultIconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file
        """
        self._dictionary = scipy.io.loadmat(self._fileName)


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._dictionary = None



class IdlSaveFileRti(MappingRti):
    """ Can read data from an IDL 'save file'.

        Uses scipy.io.readsav to read the file. This reads all data at once when the file
        is open (in contrast to lazy loading each node separately). It therefor may take a while
        to read a large save-file. During this time the application is not responsive!
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_SCIPY

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an MappingRti with None as underlying dictionary.
        """
        super(IdlSaveFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                             iconColor=self._defaultIconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file
        """
        self._dictionary = scipy.io.readsav(self._fileName)


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._dictionary = None



class WavFileRti(ArrayRti):
    """ Can read data from an WAV file.

        Uses scipy.io.wavfile.read to read the file. Tries first to read with memory mapping, if
        that fails the file is tried to be read without memory mapping(
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_SCIPY

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(WavFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                         iconColor=self._defaultIconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file.
        """
        try:
            rate, data = scipy.io.wavfile.read(self._fileName, mmap=True)
        except Exception as ex:
            logger.warning(ex)
            logger.warning("Unable to read wav with memmory mapping. Trying without now.")
            rate, data = scipy.io.wavfile.read(self._fileName, mmap=False)

        self._array = data
        self.attributes['rate'] = rate


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None
        self.attributes = {}


    def _fetchAllChildren(self):
        """ Adds an ArrayRti per column as children so that they can be inspected easily
        """
        childItems = []
        if self._array.ndim == 2:
            _nRows, nCols = self._array.shape if self._array is not None else (0, 0)
            for col in range(nCols):
                colItem = SliceRti(self._array[:, col], nodeName="channel-{}".format(col),
                                   fileName=self.fileName, iconColor=self.iconColor,
                                   attributes=self.attributes)
                childItems.append(colItem)
        return childItems




#!/usr/bin/env python
""" Convenience script that starts argos in the development environment.

    Note that this is NOT the script that will be installed by setup.py. The setup.py install
    command will create and install a dedicated argos launcher script (as specified in the
    entry_points directive).
"""
import sys, os.path

# Add the parent directory to the system path so that the package can be imported.
scriptDir = os.path.dirname(os.path.realpath(__file__))
parentDir = os.path.realpath(os.path.join(scriptDir, '..'))

sys.path.insert(0, parentDir)
from argos.main import main
main()



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the Config and GroupCtiEditor classes
"""
import logging

from argos.config.abstractcti import AbstractCti
from argos.config.groupcti import GroupCtiEditor
from argos.qt import Qt

logger = logging.getLogger(__name__)



class  BoolCti(AbstractCti):
    """ Config Tree Item to store an boolean. It can be edited using a check box
    """
    def __init__(self, nodeName, defaultData, expanded=True,
                 childrenDisabledValue=False):
        """ Constructor. For the parameters see the AbstractCti constructor documentation.

            If the data is equal to childrenDisabledValue, the node's children will be
            disabled.

            (The data will always be a bool so if childrenDisabledValue == None,
            the children will never be disabled.)
        """
        super(BoolCti, self).__init__(nodeName, defaultData, expanded=expanded)
        self.childrenDisabledValue = childrenDisabledValue


    @property
    def debugInfo(self):
        """ Returns a string with debugging information
        """
        return "{} (enabled={}, childDisabledVal={})".format(self.configValue, self.enabled,
                                                             self.childrenDisabledValue)


    def _enforceDataType(self, data):
        """ Converts to bool so that self.data always is of that type.
        """
        return bool(data)


    @property
    def data(self):
        """ Returns the data of this item.
        """
        return self._data


    @data.setter
    def data(self, data):
        """ Sets the data of this item.
            Does type conversion to ensure data is always of the correct type.
        """
        # Descendants should convert the data to the desired type here
        self._data = self._enforceDataType(data)

        #logger.debug("BoolCti.setData: {} for {}".format(data, self))
        enabled = self.enabled
        self.enableBranch(enabled and self.data != self.childrenDisabledValue)
        self.enabled = enabled


    @property
    def displayValue(self):
        """ Returns empty string since a checkbox will displayed in the value-column instead.
        """
        return ""


    def insertChild(self, childItem, position=None):
        """ Inserts a child item to the current item.

            Overridden from BaseTreeItem.
        """
        childItem = super(BoolCti, self).insertChild(childItem, position=None)

        enableChildren = self.enabled and self.data != self.childrenDisabledValue
        #logger.debug("BoolCti.insertChild: {} enableChildren={}".format(childItem, enableChildren))
        childItem.enableBranch(enableChildren)
        childItem.enabled = enableChildren
        return childItem


    @property
    def valueColumnItemFlags(self):
        """ Returns Qt.ItemIsUserCheckable so that a check box will be drawn in the config tree.
            Note that the flags include Qt.ItemIsEditable; this makes the reset button will appear.
        """
        return Qt.ItemIsUserCheckable | Qt.ItemIsEditable


    @property
    def checkState(self):
        """ Returns Qt.Checked or Qt.Unchecked.
        """
        if self.data is True:
            return Qt.Checked
        elif self.data is False:
            return Qt.Unchecked
        else:
            raise ValueError("Unexpected data: {!r}".format(self.data))


    @checkState.setter
    def checkState(self, checkState):
        """ Sets the data to given a Qt.CheckState (Qt.Checked or Qt.Unchecked).
        """
        if checkState == Qt.Checked:
            logger.debug("BoolCti.checkState setting to True")
            self.data = True
        elif checkState == Qt.Unchecked:
            logger.debug("BoolCti.checkState setting to False")
            self.data = False
        else:
            raise ValueError("Unexpected check state: {!r}".format(checkState))


    def enableBranch(self, enabled):
        """ Sets the enabled member to True or False for a node and all it's children
        """
        self.enabled = enabled

        # Disabled children and further descendants
        enabled = enabled and self.data != self.childrenDisabledValue

        for child in self.childItems:
            child.enableBranch(enabled)


    def createEditor(self, delegate, parent, _option):
        """ Creates a hidden widget so that only the reset button is visible during editing.
            :type option: QStyleOptionViewItem
        """
        return GroupCtiEditor(self, delegate, parent=parent)




class BoolGroupCti(AbstractCti):
    """ Config Tree Item to store a nullable boolean (True, False or None).

        It can be edited using a tri-state check box. However, the user can not set the value
        to partially checked directly, it will only be partially checked if some of its children
        are checked and others are not! This is a bug/behavior of Qt that won't be fixed.
        See: https://bugreports.qt.io/browse/QTBUG-7674 and
             http://comments.gmane.org/gmane.comp.lib.qt.general/925
    """
    def __init__(self, nodeName, defaultData=None, expanded=True):
        """ Constructor. For the parameters see the AbstractCti constructor documentation.
        """
        super(BoolGroupCti, self).__init__(nodeName, defaultData, expanded=expanded)


    def _enforceDataType(self, data):
        """ Converts to bool so that self.data always is of that type.
        """
        return None if data is None else bool(data)


    @property
    def displayValue(self):
        """ Returns empty string since a checkbox will displayed in the value column instead.
        """
        return ""


    @property
    def valueColumnItemFlags(self):
        """ Returns Qt.ItemIsUserCheckable so that a check box will be drawn in the config tree.
            Note that the flags include Qt.ItemIsEditable; this makes the reset button will appear.
        """
        return Qt.ItemIsTristate | Qt.ItemIsUserCheckable | Qt.ItemIsEditable


    @property
    def checkState(self):
        """ Returns Qt.Checked or Qt.Unchecked if all children are checked or unchecked, else
            returns Qt.PartiallyChecked
        """
        #commonData = self.childItems[0].data if self.childItems else Qt.PartiallyChecked
        commonData = None

        for child in self.childItems:
            if isinstance(child, BoolCti):
                if commonData is not None and child.data != commonData:
                    return Qt.PartiallyChecked
                commonData = child.data

        if commonData is True:
            return Qt.Checked
        elif commonData is False:
            return Qt.Unchecked
        else:
            raise AssertionError("Please report this bug: commonData: {!r}".format(commonData))


    @checkState.setter
    def checkState(self, checkState):
        """ Sets the data to given a Qt.CheckState (Qt.Checked or Qt.Unchecked).
        """
        logger.debug("checkState setter: {}".format(checkState))
        if checkState == Qt.Checked:
            commonData = True
        elif checkState == Qt.Unchecked:
            commonData = False
        elif checkState == Qt.PartiallyChecked:
            commonData = None
            # This never occurs, see remarks above in the classes' docstring
            assert False, "This never happens. Please report if it does."
        else:
            raise ValueError("Unexpected check state: {!r}".format(checkState))

        for child in self.childItems:
            if isinstance(child, BoolCti):
                child.data = commonData


    def createEditor(self, delegate, parent, _option):
        """ Creates a hidden widget so that only the reset button is visible during editing.
            :type option: QStyleOptionViewItem
        """
        return GroupCtiEditor(self, delegate, parent=parent)




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Widgets for displaying messages
"""
from __future__ import print_function

import logging

from argos.qt import Qt, QtGui, QtWidgets
from argos.widgets.constants import MONO_FONT, FONT_SIZE

logger = logging.getLogger(__name__)


# The main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201


class MessageDisplay(QtWidgets.QWidget):
    """ Widget that shows a label and title.
        Consists of a title label and a larger message label for details.
    """
    def __init__(self, parent=None, msg="", title="Error"):
        super(MessageDisplay, self).__init__(parent)

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.titleLabel = QtWidgets.QLabel(title)
        self.titleLabel.setTextFormat(Qt.PlainText)
        self.titleLabel.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.titleLabel.setAlignment(Qt.AlignHCenter)
        self.layout.addWidget(self.titleLabel, stretch=0)

        font = QtGui.QFont()
        font.setFamily(MONO_FONT)
        font.setFixedPitch(True)
        font.setPointSize(FONT_SIZE)

        self.messageLabel = QtWidgets.QLabel(msg)
        self.messageLabel.setFont(font)
        self.messageLabel.setTextFormat(Qt.PlainText)
        self.messageLabel.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.messageLabel.setWordWrap(True)
        self.messageLabel.setAlignment(Qt.AlignTop)
        self.messageLabel.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Plain)
        self.layout.addWidget(self.messageLabel, stretch=1)


    def setError(self, msg=None, title=None):
        """ Shows and error message
        """
        if msg is not None:
            self.messageLabel.setText(msg)

        if title is not None:
            self.titleLabel.setText(title)





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" 'Class' module that contains functions that have to do with type checking, importing, etc.

"""

import logging
import numbers

import numpy as np
import numpy.ma as ma

from argos.external import six

logger = logging.getLogger(__name__)

# For al list of encodings the the standard library see.
if six.PY2:
    URL_PYTHON_ENCODINGS_DOC = "https://docs.python.org/2/library/codecs.html#standard-encodings"
else:
    URL_PYTHON_ENCODINGS_DOC = "https://docs.python.org/3/library/codecs.html#standard-encodings"


#pylint: enable=C0103



###################
# Type conversion #
###################


# def masked_to_regular_array(masked_array, replacement_value=np.nan):
#     """ Returns a copy of the masked array where masked values have been replaced with fill values.
#
#         If masked_array is a regular numpy.ndarry, the function works, meaning a copy of the array
#         is returned.
#
#         :param masked_array: numpy masked
#         :param replacement_value: replacement value (default=np.nan).
#             If None the masked_array.fill_value is used.
#         :return: numpy.ndarray with masked arrays replaced by the fill_vluae
#     """
#     return ma.filled(masked_array, fill_value=replacement_value)

#
# def replace_missing_values(array, missing_value, replacement_value):
#     """ Returns a copy of the array where the missing_values are replaced with replacement_value.
#
#         If missing_value is None, nothing is replaced, a copy of the array is returned..
#         The missing_value can be Nan or infinite, these are replaced.
#
#         If array is a masked array the masked value are replaced (masked_to_regular_array).
#     """
#     if isinstance(array, ma.MaskedArray):
#         logger.debug("^^^^^^^^^^^^^^^^^^^ mask: {}".format(array.mask))
#
#         return array
#         return masked_to_regular_array(array, replacement_value=replacement_value)
#     if missing_value is None:
#         array = np.copy(array)
#     elif np.isnan(missing_value):
#         array[np.isnan(array)] = replacement_value
#     else:
#         array[array == missing_value] = replacement_value
#
#     return array

def environment_var_to_bool(env_var):
    """ Converts an environment variable to a boolean

        Returns False if the environment variable is False, 0 or a case-insenstive string "false"
        or "0".
    """

    # Try to see if env_var can be converted to an int
    try:
        env_var = int(env_var)
    except ValueError:
        pass

    if isinstance(env_var, numbers.Number):
        return bool(env_var)
    elif is_a_string(env_var):
        env_var = env_var.lower().strip()
        if env_var in "false":
            return False
        else:
            return True
    else:
        return bool(env_var)


def fill_values_to_nan(masked_array):
    """ Replaces the fill_values of the masked array by NaNs

        If the array is None or it does not contain floating point values, it cannot contain NaNs.
        In that case the original array is returned.
    """
    if masked_array is not None and masked_array.dtype.kind == 'f':
        check_class(masked_array, ma.masked_array)
        logger.debug("Replacing fill_values by NaNs")
        masked_array[:] = ma.filled(masked_array, np.nan)
        masked_array.set_fill_value(np.nan)
    else:
        return masked_array


# Needed because boolean QSettings in Pyside are converted incorrect the second
# time in Windows (and Linux?) because of a bug in Qt. See:
# https://www.mail-archive.com/pyside@lists.pyside.org/msg00230.html
def setting_str_to_bool(s):
    """ Converts 'true' to True and 'false' to False if s is a string
    """
    if isinstance(s, six.string_types):
        s = s.lower()
        if s == 'true':
            return True
        elif s == 'false':
            return False
        else:
            return ValueError('Invalid boolean representation: {!r}'.format(s))
    else:
        return s


#################
# Type checking #
#################

# Use '{!r}' as default float format for Python 2. This will convert the floats with repr(), which
# is necessary because str() or an empty format string will only print 2 decimals behind the point.
# In Python 3 this is not necessary: all relevant decimals are printed.
DEFAULT_NUM_FORMAT = '{!r}' if six.PY2 else '{}'

def to_string(var, masked=None, decode_bytes='utf-8', maskFormat='', strFormat='{}',
              intFormat='{}', numFormat=DEFAULT_NUM_FORMAT, noneFormat='{!r}', otherFormat='{}'):
    """ Converts var to a python string or unicode string so Qt widgets can display them.

        If var consists of bytes, the decode_bytes is used to decode the bytes.

        If var consists of a numpy.str_, the result will be converted to a regular Python string.
        This is necessary to display the string in Qt widgets.

        For the possible format string (replacement fields) see:
            https://docs.python.org/3/library/string.html#format-string-syntax

        :param masked: if True, the element is masked. The maskFormat is used.
        :param decode_bytes': string containing the expected encoding when var is of type bytes
        :param strFormat' : new style format string used to format strings
        :param intFormat' : new style format string used to format integers
        :param numFormat' : new style format string used to format all numbers except integers.
        :param noneFormat': new style format string used to format Nones.
        :param maskFormat': override with this format used if masked is True.
            If the maskFormat is empty, the format is never overriden.
        :param otherFormat': new style format string used to format all other types
    """
    #logger.debug("to_string: {!r} ({})".format(var, type(var)))

    # Decode and select correct format specifier.
    if is_binary(var):
        fmt = strFormat
        try:
            decodedVar = var.decode(decode_bytes, 'replace')
        except LookupError as ex:
            # Add URL to exception message.
            raise LookupError("{}\n\nFor a list of encodings in Python see: {}"
                              .format(ex, URL_PYTHON_ENCODINGS_DOC))
    elif is_text(var):
        fmt = strFormat
        decodedVar = six.text_type(var)
    elif is_a_string(var):
        fmt = strFormat
        decodedVar = str(var)
    elif isinstance(var, numbers.Integral):
        fmt = intFormat
        decodedVar = var
    elif isinstance(var, numbers.Number):
        fmt = numFormat
        decodedVar = var
    elif var is None:
        fmt = noneFormat
        decodedVar = var
    else:
        fmt = otherFormat
        decodedVar = var

    if maskFormat != '{}':
        try:
            allMasked = all(masked)
        except TypeError as ex:
            allMasked = bool(masked)

        if allMasked:
            fmt = maskFormat

    try:
        result = fmt.format(decodedVar)
    except Exception:
        result = "Invalid format {!r} for: {!r}".format(fmt, decodedVar)

    #if masked:
    #    logger.debug("to_string (fmt={}): {!r} ({}) -> result = {!r}".format(maskFormat, var, type(var), result))

    return result


def is_a_string(var, allow_none=False):
    """ Returns True if var is a string (ascii or unicode)

        Result             py-2  py-3
        -----------------  ----- -----
        b'bytes literal'   True  False
         'string literal'  True  True
        u'unicode literal' True  True

        Also returns True if the var is a numpy string (numpy.string_, numpy.unicode_).
    """
    return isinstance(var, six.string_types) or (var is None and allow_none)


def check_is_a_string(var, allow_none=False):
    """ Calls is_a_string and raises a type error if the check fails.
    """
    if not is_a_string(var, allow_none=allow_none):
        raise TypeError("var must be a string, however type(var) is {}"
                        .format(type(var)))


def is_text(var, allow_none=False):
    """ Returns True if var is a unicode text

        Result             py-2  py-3
        -----------------  ----- -----
        b'bytes literal'   False False
         'string literal'  False True
        u'unicode literal' True  True

        Also works with the corresponding numpy types.
    """
    return isinstance(var, six.text_type) or (var is None and allow_none)

# Not used yet
# def check_is_text(var, allow_none=False):
#     """ Calls is_text and raises a type error if the check fails.
#     """
#     if not is_text(var, allow_none=allow_none):
#         raise TypeError("var must be a text (unicode str), however type(var) is {}"
#                         .format(type(var)))


def is_binary(var, allow_none=False):
    """ Returns True if var is a binary (bytes) objects

        Result             py-2  py-3
        -----------------  ----- -----
        b'bytes literal'   True  True
         'string literal'  True  False
        u'unicode literal' False False

        Also works with the corresponding numpy types.
    """
    return isinstance(var, six.binary_type) or (var is None and allow_none)


# Not used yet
# def check_is_text(var, allow_none=False):
#     """ Calls is_binary and raises a type error if the check fails.
#     """
#     if not is_binary(var, allow_none=allow_none):
#         raise TypeError("var must be a binary (bytes), however type(var) is {}"
#                         .format(type(var)))


# Not used. Remove?
# def is_a_numpy_string(var, allow_none=False):
#     """ Returns True if var is of type: numpy.string_, numpy.unicode_
#
#         :param var: variable of which we want to know if it is a string
#         :type var: any type
#         :returns: True if var is of type string
#         :rtype: Boolean
#     """
#     return isinstance(var, (np.string_, np.unicode_)) or (var is None and allow_none)


def is_a_sequence(var, allow_none=False):
    """ Returns True if var is a list or a tuple (but not a string!)
    """
    return isinstance(var, (list, tuple)) or (var is None and allow_none)


def check_is_a_sequence(var, allow_none=False):
    """ Calls is_a_sequence and raises a type error if the check fails.
    """
    if not is_a_sequence(var, allow_none=allow_none):
        raise TypeError("var must be a list or tuple, however type(var) is {}"
                        .format(type(var)))


def is_a_mapping(var, allow_none=False):
    """ Returns True if var is a dictionary # TODO: ordered dict
    """
    return isinstance(var, dict) or (var is None and allow_none)


def check_is_a_mapping(var, allow_none=False):
    """ Calls is_a_mapping and raises a type error if the check fails.
    """
    if not is_a_mapping(var, allow_none=allow_none):
        raise TypeError("var must be a dict, however type(var) is {}"
                        .format(type(var)))


def is_an_array(var, allow_none=False):
    """ Returns True if var is a numpy array.
    """
    return isinstance(var, np.ndarray) or (var is None and allow_none)


def check_is_an_array(var, allow_none=False):
    """ Calls is_an_array and raises a type error if the check fails.
    """
    if not is_an_array(var, allow_none=allow_none):
        raise TypeError("var must be a NumPy array, however type(var) is {}"
                        .format(type(var)))


def array_is_structured(array):
    """ Returns True if the array has a structured data type.
    """
    return bool(array.dtype.names)


def array_has_real_numbers(array):
    """ Uses the dtype kind of the numpy array to determine if it represents real numbers.

        That is, the array kind should be one of: i u f

        Possible dtype.kind values.
            b 	boolean
            i 	signed integer
            u 	unsigned integer
            f 	floating-point
            c 	complex floating-point
            m 	timedelta
            M 	datetime
            O 	object
            S 	(byte-)string
            U 	Unicode
            V 	void
    """
    kind = array.dtype.kind
    assert kind in 'biufcmMOSUV', "Unexpected array kind: {}".format(kind)
    return kind in 'iuf'


def check_class(obj, target_class, allow_none = False):
    """ Checks that the  obj is a (sub)type of target_class.
        Raises a TypeError if this is not the case.

        :param obj: object whos type is to be checked
        :type obj: any type
        :param target_class: target type/class
        :type target_class: any class or type
        :param allow_none: if true obj may be None
        :type allow_none: boolean
    """
    if not isinstance(obj, target_class):
        if not (allow_none and obj is None):
            raise TypeError("obj must be a of type {}, got: {}"
                            .format(target_class, type(obj)))


#############
# Type info #
#############

# TODO: get_class_name and type_name the same? Not for old style classes
# #http://stackoverflow.com/questions/1060499/difference-between-typeobj-and-obj-class

def type_name(var):
    """ Returns the name of the type of var"""
    return type(var).__name__


def get_class_name(obj):
    """ Returns the class name of an object.
    """
    return obj.__class__.__name__


def get_full_class_name(obj):
    """ Returns the full class name of an object. This includes packages and module names.

        It depends on where the class is imported so only use for testing and debugging!
    """
    return "{}.{}".format(obj.__class__.__module__, obj.__class__.__name__)

#############
# Importing #
#############

def import_symbol(full_symbol_name):
    """ Imports a symbol (e.g. class, variable, etc) from a dot separated name.
        Can be used to create a class whose type is only known at run-time.

        The full_symbol_name must contain packages and module,
        e.g.: 'argos.plugins.rti.ncdf.NcdfFileRti'

        If the module doesn't exist an ImportError is raised.
        If the class doesn't exist an AttributeError is raised.
    """
    parts = full_symbol_name.rsplit('.', 1)
    if len(parts) == 2:
        module_name, symbol_name = parts
        module_name = str(module_name) # convert from possible unicode
        symbol_name = str(symbol_name)
        #logger.debug("From module {} importing {!r}".format(module_name, symbol_name))
        module = __import__(module_name, fromlist=[symbol_name])
        cls = getattr(module, symbol_name)
        return cls
    elif len(parts) == 1:
        # No module part, only a class name. If you want to create a class
        # by using name without module, you should use globals()[symbol_name]
        # We cannot do this here because globals is of the module that defines
        # this function, not of the modules where this function is called.
        raise ImportError("full_symbol_name should contain a module")
    else:
        assert False, "Bug: parts should have 1 or elements: {}".format(parts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import

# To make a release follow these steps:
#   python setup.py sdist
#   twine upload dist/argos-0.2.0rc1.tar.gz

# See also https://packaging.python.org/en/latest/distributing.html
# TODO: still can't make a wheel even following the instructions in the link below.
# http://stackoverflow.com/questions/26664102/why-can-i-not-create-a-wheel-in-pyt

import sys

def err(*args, **kwargs):
    sys.stderr.write(*args, **kwargs)
    sys.stderr.write('\n')


try:
    from setuptools import setup, find_packages
except ImportError:
    err("Argos requires setuptools for intallation. (https://pythonhosted.org/an_example_pypi_project/setuptools.html)")
    err("You can download and install it simply with: python argos/external/ez_setup.py")
    sys.exit(1)


from argos import info

assert not info.DEBUGGING, "info.DEBUGGING must be False to make a release."

if sys.version_info < (2,7) or ((3,0) <= sys.version_info < (3, 4)):
    err("Arogs requires Python 2.7 and higher or 3.4 and higher.")
    sys.exit(1)


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

# From http://setuptools.readthedocs.io/en/latest/setuptools.html?highlight=install_requires#declaring-dependencies
# Using the install_requires option has three effects:
#  1) When your project is installed, either by using EasyInstall, setup.py install, or
#     setup.py develop, all of the dependencies not already installed will be located (via PyPI),
#     downloaded, built (if necessary), and installed.
#  2) Any scripts in your project will be installed with wrappers that verify the availability of
#     the specified dependencies at runtime, and ensure that the correct versions are added to
#     sys.path (e.g. if multiple versions have been installed).
#  3) Python Egg distributions will include a metadata file listing the dependencies.
#
# The reason for disabling install_requires is that, if a dependency is installed without pip or
# easy_install, the argos launcher script that is generated by setup.py gives the following error:
# pkg_resources.DistributionNotFound: The 'PyQt5' distribution was not found and is required by argos
# It was most likely caused by not having pip installed but it is unclear how to debug this and
# google didn't help. I don't know how to resolve these errors, so to save myself a lot of headaches
# I disabled automatic dependency installation.

# Declaring depdencies. DISABLED, see comment above.
install_requires = [
    "PyQt5 >= 5.6.0", # Don't know which version is minimal
    "numpy >= 1.11",
    # Argos will technically work without pyqtgraph and h5py, but with very limited functionality.
    "pyqtgraph >= 0.10",
    "h5py >= 2.6"
]


setup(
    name = info.REPO_NAME,
    version = info.VERSION,
    description = info.SHORT_DESCRIPTION,
    long_description = readme + '\n\n' + history,
    author = info.AUTHOR,
    author_email = info.EMAIL,
    license = "GPLv3",
    url=info.PROJECT_URL,
    packages = find_packages(),
    package_data = {'': ['HISTORY.rst'], info.PACKAGE_NAME: ['img/snipicons/*']},
    entry_points={'gui_scripts': ['argos = argos.main:main']},
    #install_requires = install_requires, # DISABLED. See coments above.
    zip_safe = False,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: Qt',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities',
    ],
    keywords = 'NCDF HDF plotting graphs',
    #test_suite='tests',
    #tests_require=test_requirements
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import unittest, logging, sys, copy
from json import loads, dumps

from argos import configBasicLogging
from argos.qt import QtWidgets
from argos.config.abstractcti import ctiDumps, CtiDecoder
from argos.config.untypedcti import UntypedCti
from argos.config.qtctis import ColorCti



class TestUntypedCtis(unittest.TestCase):

    def setUp(self):
        self.invisibleRootItem = UntypedCti(nodeName='<invisible-root>', defaultData=0.123456789012345678901234567890)
        self.invisibleRootItem.insertChild(UntypedCti(nodeName='kid', defaultData=-7))


    def test__eq__(self):

        ctiIn = UntypedCti('parent', defaultData=7)
        self.assertEqual(ctiIn, ctiIn)
        self.assertEqual(ctiIn, UntypedCti('parent', defaultData=7))

        self.assertNotEqual(ctiIn, UntypedCti('parent', defaultData=9))

        ctiOut = UntypedCti('parent', defaultData=7)
        ctiIn.insertChild(UntypedCti('kid', defaultData=23))
        self.assertNotEqual(ctiIn, ctiOut)

        ctiOut.insertChild(UntypedCti('kid', defaultData=23))
        self.assertEqual(ctiIn, ctiOut)

        ctiIn.childItems[0].data = 99
        self.assertNotEqual(ctiIn, ctiOut)


    # This functionality is currently not used? Depricated?
    def __not_used__testJson(self):

        # encoding
        jstr = ctiDumps(self.invisibleRootItem)
        #print(jstr)

        # decoding
        #testItem = json.loads(jstr, object_hook=jsonAsCti)
        testItem = loads(jstr, cls=CtiDecoder)
        #print(mydumps(testItem))

        self.assertEqual(testItem, self.invisibleRootItem)


    def testNonDefaults(self):

        # A data different than its default should be restored
        ctiIn  = UntypedCti('parent', defaultData=7)
        defDict = ctiIn.getNonDefaultsDict()
        #print(defDict)
        ctiOut = UntypedCti('parent', defaultData=7)
        ctiOut.setValuesFromDict(defDict)
        self.assertEqual(ctiIn, ctiOut)

        # A data is not written if it's the default and
        # is read correctly if the default stays the same
        ctiIn  = UntypedCti('parent', defaultData=7)
        defDict = ctiIn.getNonDefaultsDict()
        #print(defDict)
        ctiOut = UntypedCti('parent', defaultData=7)
        ctiOut.setValuesFromDict(defDict)
        self.assertEqual(ctiIn, ctiOut)

        # A default data should have no effect it the default changes in the future.
        ctiIn  = UntypedCti('parent', defaultData=7)
        defDict = ctiIn.getNonDefaultsDict()
        #print(defDict)
        ctiOut = UntypedCti('parent', defaultData=9)
        ctiOut.setValuesFromDict(defDict)
        self.assertNotEqual(ctiIn, ctiOut)

        # Test children
        ctiIn  = UntypedCti('parent', defaultData=6)
        ctiIn.insertChild(UntypedCti('child1', defaultData=1))
        ctiIn.insertChild(UntypedCti('child2', defaultData=2))
        defDict = ctiIn.getNonDefaultsDict()
        #print(defDict)
        ctiOut = UntypedCti('parent', defaultData=6)
        ctiOut.insertChild(UntypedCti('child1', defaultData=1))
        ctiOut.insertChild(UntypedCti('child2', defaultData=2))
        ctiOut.setValuesFromDict(defDict)
        self.assertEqual(ctiIn, ctiOut)


    def tearDown(self):
        pass



class TestSimpleCtis(unittest.TestCase):

    def setUp(self):
        self.invisibleRootItem = UntypedCti(nodeName='<invisible-root>', defaultData=0.123456789012345678901234567890)
        self.invisibleRootItem.insertChild(UntypedCti(nodeName='kid', defaultData=-7))

    def tearDown(self):
        pass


    def testColorCti(self):

        colorStr = '#FF33EE'
        cti = ColorCti('color', defaultData=colorStr)
        self.assertEqual(cti.data, QtGui.QColor(colorStr))
        self.assertEqual(cti.data, QtGui.QColor(colorStr))
        self.assertEqual(cti.displayValue, colorStr)


    def closedLoop(self, ctiIn):
        """ serialize cti default values to json and back
        """
        nonDefaults = ctiIn.getNonDefaultsDict()
        json = dumps(nonDefaults)
        valuesDict = loads(json)

        ctiOut = copy.deepcopy(ctiIn)
        ctiOut._data = None
        ctiOut.setValuesFromDict(valuesDict)
        return ctiOut


    def testClosedLoop(self):
        # A data different than its default should be restored
        ctiIn  = ColorCti('parent', defaultData='#ABCDEF')
        ctiIn.data='#DEADBE' # works only if data is set, otherwise ctiOut.data will be None
        ctiOut = self.closedLoop(ctiIn)
        print("ctiIn {}".format(ctiIn.__dict__))
        print("ctiOut {}".format(ctiOut.__dict__))
        self.assertEqual(ctiIn, ctiOut)


if __name__ == '__main__':
    configBasicLogging(level='DEBUG')
    unittest.main()





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" RTI properties inspector.
"""
import logging

from argos.qt import QtWidgets, QtCore
from argos.repo.detailpanes import DetailTablePane
from argos.repo.repotreemodel import RepoTreeModel

logger = logging.getLogger(__name__)



class PropertiesPane(DetailTablePane):
    """ Shows the properties of the selected repo tree item.

        The properties correspond to a column in the repository tree view but typically those
        columns are hidden to save screen space. That's why this details pane is useful.
    """
    _label = "Properties"

    HEADERS = ["Name", "Value"]
    (COL_PROP_NAME, COL_VALUE) = range(len(HEADERS))

    def __init__(self, repoTreeView, parent=None):
        super(PropertiesPane, self).__init__(repoTreeView, parent=parent)
        self.table.addHeaderContextMenu(enabled = {'Name': False, 'Value': False}) # disable action

        self.table.setTextElideMode(QtCore.Qt.ElideMiddle)

        tableHeader = self.table.horizontalHeader()
        tableHeader.resizeSection(self.COL_PROP_NAME, 125)
        tableHeader.resizeSection(self.COL_VALUE, 150)


    def _drawContents(self, currentRti=None):
        """ Draws the attributes of the currentRTI
        """
        table = self.table
        table.setUpdatesEnabled(False)
        try:
            table.clearContents()
            verticalHeader = table.verticalHeader()
            verticalHeader.setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

            if currentRti is None:
                return

            # Each column in the repo tree corresponds to a row in this detail pane.
            repoModel = self._repoTreeView.model()
            propNames = RepoTreeModel.HEADERS
            table.setRowCount(len(propNames))

            for row, propName in enumerate(propNames):
                nameItem = QtWidgets.QTableWidgetItem(propName)
                nameItem.setToolTip(propName)
                table.setItem(row, self.COL_PROP_NAME, nameItem)
                propValue = repoModel.itemData(currentRti, row)
                propItem = QtWidgets.QTableWidgetItem(propValue)
                propItem.setToolTip(propValue)
                table.setItem(row, self.COL_VALUE, propItem)
                table.resizeRowToContents(row)

            verticalHeader.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        finally:
            table.setUpdatesEnabled(True)






# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Uses the Python Imaging Library (Pillow) to open an image
"""
import logging
import numpy as np

from PIL import Image

from argos.info import DEBUGGING
from argos.repo.iconfactory import RtiIconFactory
from argos.repo.memoryrtis import ArrayRti, SliceRti

logger = logging.getLogger(__name__)

ICON_COLOR_PILLOW = '#880088'


class PillowBandRti(SliceRti):
    """ Image band repo tree item. Will typically be a child of a PillowFileRti

        No dedicated constructor defined (reuses the ArrayRti constructor)
    """

    @property
    def dimensionNames(self):
        """ Returns ['Y', 'X'].
            The underlying array is expected to be 2-dimensional. If this is not the case we fall
            back on the default dimension names ['Dim-0', 'Dim-1', ...]
        """
        if self._array is None:
            return []

        if self._array.ndim != 2:
            # Defensive programming: fall back on default names
            msg = "Expected 2D image. Got: {}".format(self._array.ndim)
            if DEBUGGING:
                raise ValueError(msg)
            logger.warn(msg)
            return super(PillowBandRti, self).dimensionNames
        else:
            return ['Y', 'X']



class PillowFileRti(ArrayRti):
    """ Opens an image file with the Python Imaging Library (Pillow)

        See https://python-pillow.org/
    """
    _defaultIconGlyph = RtiIconFactory.FILE
    _defaultIconColor = ICON_COLOR_PILLOW

    def __init__(self, nodeName='', fileName=''):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(PillowFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                            iconColor=self._defaultIconColor)
        self._checkFileExists()
        self._bands = [] # image band names


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses  open the underlying file
        """
        with Image.open(self._fileName) as image:
            self._array = np.asarray(image)
            self._bands = image.getbands()

            # Fill attributes. For now assume that the info item are not overridden by
            # the Image items.
            self._attributes = dict(image.info)
            self._attributes['Format'] = image.format
            self._attributes['Mode'] = image.mode
            self._attributes['Size'] = image.size
            self._attributes['Width'] = image.width
            self._attributes['Height'] = image.height


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None
        self._bands = []
        self._attributes = {}


    def _fetchAllChildren(self):
        """ Adds the bands as separate fields so they can be inspected easily.
        """
        bands = self._bands
        if len(bands) != self._array.shape[-1]:
            logger.warn("No bands added, bands != last_dim_lenght ({} !: {})"
                        .format(len(bands), self._array.shape[-1]))
            return []

        childItems = []
        for bandNr, band in enumerate(bands):
            bandItem = PillowBandRti(self._array[..., bandNr],
                                     nodeName=band, fileName=self.fileName,
                                     iconColor=self.iconColor, attributes=self._attributes)
            childItems.append(bandItem)
        return childItems


    @property
    def attributes(self):
        """ The attribute dictionary. Con
        """
        return self._attributes


    @property
    def dimensionNames(self):
        """ Returns ['Y', 'X', 'Band'].
            The underlying array is expected to be 3-dimensional. If this is not the case we fall
            back on the default dimension names ['Dim-0', 'Dim-1', ...]
        """
        if self._array is None:
            return []

        if self._array.ndim == 2:
            return ['Y', 'X']
        elif self._array.ndim == 3:
            return ['Y', 'X', 'Band']
        else:
            # Defensive programming: fall back on default names
            msg = "Expected 3D image. Got: {}".format(self._array.ndim)
            if DEBUGGING:
                raise ValueError(msg)
            logger.warn(msg)
            return super(PillowFileRti, self).dimensionNames




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_argos
----------------------------------

Tests for `argos` module.
"""
import sys
import unittest

sys.path.append('..')
#print(sys.path)

from argos.qt import QtCore, getQApplicationInstance, USE_PYQT


class TestArgos(unittest.TestCase):

    def setUp(self):
        self.app = getQApplicationInstance()

        self.groupName = '__test__'
        self.qs = QtCore.QSettings()
        self.qs.remove(self.groupName) # start with clean slate
        self.qs.beginGroup(self.groupName)


    def test_read_write(self):

        self.qs.setValue('int', -6)
        self.assertEqual(-6, self.qs.value('int'))

        self.qs.setValue('float', 7.7)
        self.assertEqual(7.7, self.qs.value('float'))

        self.qs.setValue('str', 'six')
        self.assertEqual('six', self.qs.value('str'))

        large_str = '0123456789' * 10000
        self.qs.setValue('large_str', large_str)
        self.assertEqual(large_str, self.qs.value('large_str'))

        for p in range(5):
            n = pow(10, p)
            arr = range(n)
            name = 'arr10pow{}'.format(p)
            self.qs.setValue(name, arr)
            print("\np={}, n={}".format(p, n))
            print(repr(arr))
            print(repr(self.qs.value(name)))
            self.assertEqual(arr, arr)



    def tearDown(self):
        self.qs.endGroup()
        self.qs.remove(self.groupName)


if __name__ == '__main__':
    print ('Using: {}'.format('PyQt' if USE_PYQT else 'PySide'))
    unittest.main()

